use rusttype;

use std::collections::HashMap;

use ::{FontAtlas, GlyphMetadata};
use ::geometry::Rect;
use ::texture::{Buffer2d, Bitmap};
use ::packer::SkylinePacker;

pub struct Font<'a> {
    font: rusttype::Font<'a>
}

#[allow(dead_code)]
impl<'a> Font<'a> {
    /// Construct a Textium font from a RustType one.
    pub fn new(rt_font: rusttype::Font<'a>) -> Font {
        Font {
            font: rt_font
        }
    }


    /// Build a font atlas for a set of characters in a font, rasterized at a given scale.
    ///
    /// `margin` is the distance between glyphs in the atlas, in pixels.
    /// `width` and `height` are the starting size of the bitmap.
    ///
    /// The packer may enlarge the bitmap beyond its initial dimensions in order to fit all glyphs.
    pub fn make_atlas<I>(&self, chrs: I, scale: f32, margin: u32, width: usize, height: usize)
        -> (FontAtlas, Bitmap<u8>, f32)
            where I: Iterator<Item=char>
    {
        let mut atlas = FontAtlas {glyph_data: HashMap::new()};
        let mut packer = SkylinePacker::new(Bitmap::new(width, height));
        packer.set_margin(margin);
        for chr in chrs {
            if let Some((mut info, rendered)) = self.rasterize_char(chr, scale) {
                let r = packer.pack_resize(&rendered);
                info.bounding_box = r;
                atlas.glyph_data.insert(chr, info);
            } else if chr == ' ' {
                // generate a blank pixel for the space texture
                let (mut info, _) = self.rasterize_char(' ', scale).unwrap();
                let empty_bitmap = Bitmap::new(1, 1);
                let r = packer.pack_resize(&empty_bitmap, |(w, h)| (w*2, h*2));
                info.bounding_box = r;
                atlas.glyph_data.insert(chr, info);
            } else {
                // TODO: use a Result instead of panic!ing
                panic!("Cannot render character {}", chr);
            }
        }

        let v_metrics = self.font.v_metrics(rusttype::Scale::uniform(scale));
        let line_height = v_metrics.ascent + v_metrics.descent + v_metrics.line_gap;

        (atlas, packer.into_buf(), line_height)
    }

    pub fn rasterize_char(&self, chr: char, scale: f32)
        -> Option<(GlyphMetadata, Bitmap<u8>)>
    {
        let glyph = match self.font.glyph(chr) {
            Some(g) => g,
            None => return None, // propogate missing glyphs out of the function
        }.scaled(rusttype::Scale::uniform(scale));

        let h_metrics = glyph.h_metrics();
        let glyph = glyph.positioned(rusttype::Point {x: 0., y: 0.});
        let bbox = match glyph.pixel_bounding_box() {
            Some(b) => b,
            None => return None, // if the bounding box is missing, we can't rasterize
        };

        let mut out: Bitmap<u8> = Bitmap::new(bbox.width() as usize, bbox.height() as usize);
        glyph.draw(|x, y, v| {
            out.set(x as usize, y as usize, (v * 255.0) as u8);
        });

        println!("{:?}", bbox);

        let data = GlyphMetadata {
            scale: scale,
            bounding_box: Rect {
                x: bbox.min.x as usize,
                y: bbox.min.y as usize,
                w: bbox.width() as usize,
                h: bbox.height() as usize,
            },

            advance_width: h_metrics.advance_width,
            left_bearing: h_metrics.left_side_bearing,
        };

        Some((data, out))
    }
}
